from typing import TypeVar, NamedTuple, Callable, Optional

T = TypeVar("T")


class DependencyInfo(NamedTuple):
    cls: type
    factory: Callable
    store: bool = True
    kwargs: Optional[dict] = None


class TypeHintInfo(NamedTuple):
    collection: type
    inner_type: type



import collections
import typing as t

from .types import T, TypeHintInfo


def is_generic_type_hint(hint):
    return hasattr(hint, "__origin__")


def parse_type_hint(hint) -> t.Optional[TypeHintInfo]:
    """
    Get the constructor for iterable/sequence type hints:
    returns the concrete type belonging to the type hint, ie `set` for `typing.Set`
    for the abstract type hints typing.Iterable
    :param hint: A Generic Type hint
    :returns: TypeHintInfo, or None if it could not be parsed into a list-like collection of
    `type`. eg for nested generic types (`List[List[int]]`)
    """
    # hint must be a Type hint of type Iterable, but not a Mapping
    if (
        not is_generic_type_hint(hint)
        or hint.__origin__ is None  # py36
        or not issubclass(hint.__origin__, collections.abc.Iterable)
        or issubclass(hint.__origin__, collections.abc.Mapping)
    ):
        return None

    collection_type = hint.__origin__

    if issubclass(collection_type, tuple):
        # Must be variable length tuple
        if len(hint.__args__) != 2 or hint.__args__[1] != Ellipsis:
            return None

    # abstract types like Iterable and Sequence are given as list
    if collection_type in vars(collections.abc).values():
        collection_type = list

    inner_type = getattr(hint, "__args__", [None])[0]
    if isinstance(inner_type, t.ForwardRef):
        inner_type = inner_type.__forward_arg__
    elif not isinstance(inner_type, (str, type)):
        return None

    return TypeHintInfo(collection_type, inner_type)


class _Stack(list, t.List[T]):
    def push(self, item):
        self.append(item)
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pop()


class _LookupStack(_Stack):
    def __str__(self):
        return " -> ".join(getattr(i, "__name__", str(i)) for i in self)


EMPTY = object()



from __future__ import annotations
from gimme.helpers import EMPTY
import typing as t

if t.TYPE_CHECKING:
    from gimme.repository import LayeredRepository


class Attribute:
    name: str

    def __init__(self, cls_or_name, repo: LayeredRepository, lazy=True):
        self.dependency = cls_or_name
        self.lazy = lazy
        self.repo = repo

    def __get__(self, instance, owner):
        if instance is None:
            return self

        obj = instance.__dict__.get(self.name, EMPTY)
        if obj is EMPTY:
            obj = self.repo.get(self.dependency)
            instance.__dict__[self.name] = obj
        return obj

    def __set_name__(self, owner, name):
        self.name = name



import typing as t

from gimme.repository import LayeredRepository, SimpleRepository
from gimme.attribute import Attribute
from gimme.resolvers import Resolver, TypeHintingResolver
from gimme.types import DependencyInfo, T


def that(kind: t.Union[t.Type[T], str, callable], **kwargs) -> T:
    """Request an object from the :class:`Repository <gimme.repository.LayeredRepository>` by type.
    If it does not exist in the :class:`Repository <gimme.repository.LayeredRepository>` yet,
    it will be created, and any dependencies of the class will be resolved recursively. If the
    class has been registered with a specific factory function, that function will be used to
    create the object.

    Alternatively, a callable, such as function, may be given as the kind. In this case,
    ``gimme.that`` tries to resolve all input parameters of the function, runs the function with
    the input parameters and returns the result. In this case the result is never stored in the
    :class:`Repository <gimme.repository.LayeredRepository>`. However, any dependencies that are
    created when resolving the input parameters, may be stored.

    Optionally, if the class has been registered using the :func:`gimme.dependency` decorator or
    :func:`gimme.register` function, the object type may also be specified using class name as a
    string.

    :param kind: either a class, identified by the type-object itself or a string, or a
        ``callable``

    :rtype: ~T
    """
    return current_repo().get(kind, kwargs=kwargs or None)


def later(cls_or_str: t.Union[type, str], lazy: bool = True) -> "Attribute":
    """Use this as a descriptor in your class definition when dealing with circular dependencies.
    Normally, having circular dependencies would prevent instantiating these classes.

    By default, using this descriptor to specify a dependency defers the resolution of this
    dependency until the first time it is used.

    :param cls_or_str: The type of which to retrieve an object
    :param lazy: When set to ``False``, the dependency is resolved immediately the first time the
        dependent class is instantiated. This precludes the use of ``gimme.later`` for resolving
        circular dependencies, but instead now it works the same way of using type annotations for
        specifying dependencies. If you don't like using type annotations, you may use
        ``gimme.later`` with ``lazy=False`` to get the same behaviour.
    """
    return Attribute(cls_or_str, repo=current_repo(), lazy=lazy)


def add(obj: t.Any, deep: bool = True):
    """Add an object to the :class:`Repository <gimme.repository.LayeredRepository>` after which
    it will be available for dependent classes when resolving their dependencies.

    By default, the object will be registered for both its own class as well as any subclasses (up
    to and including ``object``). Setting ``deep=False`` switches this off, and the object will
    only be registered for its own class

    :param obj: The object to add to the :class:`Repository <gimme.repository.LayeredRepository>`
    :param deep: Whether to also register `obj` for all base classes of `obj`
    """
    return current_repo().add(obj, deep=deep)


def dependency(cls: t.Type[T]) -> t.Type[T]:
    """
    .. deprecated:: 0.1.2
        Use :func:`gimme.register` instead.
    """
    current_repo().register(cls)
    return cls


def register(
    cls: type = None,
    factory: t.Optional[t.Callable] = None,
    info: t.Optional[DependencyInfo] = None,
    store: bool = True,
    kwargs: t.Optional[dict] = None,
):
    """
    Register a class in the :class:`Repository <gimme.repository.LayeredRepository>`. This makes
    the class available for identification by string (see :func:`gimme.that` and
    :func:`gimme.later`) and allows for additional control over class instantiation.

    Alternatively, instead of giving a class (and optional parameters), you can also give a
    :class:`~gimme.types.DependencyInfo` object directly that contains the configuration.

    :param cls: The class to register in the repository.
    :param factory: Optional factory to create an instance, if not specified, use the class'
        constructor instead (ie. ``cls.__init__``)
    :param info: An object that contains the class instantiation configuration. Only provide this
        when no ``cls`` parameter is given
    :param store: Whether to store an instantiated class in the
        :class:`Repository <gimme.repository.LayeredRepository>`. The default effectively makes
        every class a singleton, but if you want to change this behaviour, set ``store=False``
    :param kwargs: Additional keyword arguments to pass to the factory (or ``cls.__init__``) for
        construction
    """
    current_repo().register(cls, factory, info=info, store=store, kwargs=kwargs)


def setup(
    objects: t.Optional[t.Iterable[t.Any]] = None,
    types: t.Optional[t.Iterable[t.Union[type, DependencyInfo]]] = None,
    resolvers: t.Optional[t.Iterable[Resolver]] = None,
):
    """Setup and configure your :class:`Repository <gimme.repository.LayeredRepository>`

    :param objects: Any objects to add to the
        :class:`Repository <gimme.repository.LayeredRepository>`
    :param types: Any types / classes to register
    :param resolvers: Any additional :class:`Resolvers <gimme.resolvers.Resolver>` to add for
        dependency resolution
    """
    if objects is not None:
        for obj in objects:
            add(obj)

    if types is not None:
        for tp in types:
            if isinstance(tp, DependencyInfo):
                register(info=tp)
            else:
                register(cls=tp)
    if resolvers is not None:
        for resolver in resolvers:
            add_resolver(resolver)


def add_resolver(resolver: Resolver):
    """Add a custom :class:`~gimme.resolvers.Resolver` to the
    :class:`Repository <gimme.repository.LayeredRepository>`
    """
    current_repo().add_resolver(resolver)


def current_repo() -> LayeredRepository:
    return _repository.get(LayeredRepository)


def context():
    """return a context manager that adds a repository to the stack and yields it"""
    repo = current_repo()
    resolvers = repo.resolvers.copy()
    return repo.push(SimpleRepository(resolvers))


def pop_context():
    """Pops the current repository from the repository stack, effectively resetting the state of
    the repository to a previous state"""
    repo = current_repo()
    return repo.pop()


_repository = SimpleRepository(resolvers=[TypeHintingResolver()])
_repository.add(TypeHintingResolver())

# Some aliases for professional (ie. boring) people
get = that
attribute = later



class CannotResolve(TypeError):
    """Raised when a :class:`~gimme.resolvers.Resolver` cannot resolve the dependencies of a class,
    and therefore cannot provide an instance
    """

    pass


class PartiallyResolved(RuntimeError):
    """Raised when an :class:`~gimme.resolvers.Resolver` did do some work on resolving
    dependencies, but not enough to be able to instantiate the class
    """

    pass


class CircularDependency(RuntimeError):
    """Raised when a circular dependency is detected and the requested class cannot be
    instantiated"""

    pass



import re
from pathlib import Path

from setuptools import setup, find_packages

CURRENT_DIR = Path(__file__).parent


def get_long_description() -> str:
    readme = CURRENT_DIR / "README.rst"
    with open(readme) as file:
        return file.read()


def find_version(fname, var_name="__version__") -> str:
    """Attempts to find the version number in the file names fname.
    Raises RuntimeError if not found.
    """
    version = None
    regex = re.compile(rf'{var_name} = [\'"]([^\'"]*)[\'"]')

    with open(fname, "r") as file:
        for line in file:
            match = regex.match(line)
            if match:
                version = match.group(1)
                break
    if not version:
        raise RuntimeError("Cannot find version information.")
    return version


EXTRAS_REQUIRE = {
    "test": ["pytest==6.2.5", "pytest-cov==3.0.0"],
    "docs": ["sphinx==4.3.2", "sphinx-autodoc-typehints==1.15.2"],
}
EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["test"]
    + EXTRAS_REQUIRE["docs"]
    + ["pre-commit==2.16.0", "flake8==4.0.1", "tox==3.24.5"]
)


setup(
    name="gimme-that",
    description="Simple and flexible dependency injection framework.",
    long_description=get_long_description(),
    long_description_content_type="text/x-rst",
    keywords="inversion-of-control IoC DI dependency-injection",
    author="Pelle Koster",
    url="https://github.com/elfjes/gimme-that",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.6",
    extras_require=EXTRAS_REQUIRE,
    test_suite="tests",
    version=find_version("src/gimme/__init__.py"),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries",
    ],
    project_urls={"Documentation": "https://gimme-that.readthedocs.io/"},
)



from .main import (
    add,
    add_resolver,
    attribute,
    context,
    current_repo,
    dependency,
    get,
    later,
    pop_context,
    register,
    setup,
    that,
)
from .repository import LayeredRepository, SimpleRepository
from .resolvers import AttributeResolver, TypeHintingResolver
from .types import DependencyInfo

__version__ = "0.2.2"

__all__ = [
    "add",
    "add_resolver",
    "attribute",
    "context",
    "current_repo",
    "dependency",
    "get",
    "later",
    "pop_context",
    "register",
    "setup",
    "that",
    "LayeredRepository",
    "SimpleRepository",
    "AttributeResolver",
    "TypeHintingResolver",
    "DependencyInfo",
]



from __future__ import annotations

from itertools import chain
import typing as t

from .exceptions import CannotResolve, CircularDependency, PartiallyResolved
from .helpers import EMPTY, _LookupStack, _Stack
from .types import DependencyInfo, T

from .resolvers import Resolver


class SimpleRepository:
    def __init__(self, resolvers: t.Iterable[Resolver]):

        self.resolvers: t.List[Resolver] = list(resolvers)
        self.types_by_str: t.Dict[str, t.Type[T]] = {}
        self.types: t.Dict[t.Type, DependencyInfo] = {}
        self.instances: t.Dict[t.Type[T], t.List[T]] = {}
        self.lookup_stack = _LookupStack()

    def get(
        self, key: t.Union[t.Type[T], str], many=False, repo=None, kwargs=None
    ) -> t.Union[t.List[T], T]:
        if isinstance(key, str):
            key = self.types_by_str.get(key)
        if key is None:
            raise CannotResolve(f"Could not resolve for key {key}")

        if many:
            return self.instances.get(key, [])

        if not isinstance(key, type) and callable(key):
            name = getattr(key, "__name__", str(key))
            return self.resolve(factory=key, key=name, repo=repo, kwargs=kwargs)

        if kwargs is not None or key not in self.instances:
            inst = self.create(key, repo=repo, kwargs=kwargs)
            return inst

        instances = self.instances[key]
        return instances[-1]

    def create(self, key: t.Type[T], repo=None, kwargs=None) -> T:
        """Instantiate the object and all its dependencies
        :param key: The class / factory function to instantiate
        :param repo: The current ``Repository`` used for requesting dependencies
        :param kwargs: Any user specified keyword arguments (see :func:`gimme.get`). These
            will have preference over any keyword arguments this function supplies and any
            keyword arguments supplied by :func:`gimme.register`

        """
        repo = repo or self
        if not isinstance(key, type):
            raise TypeError(f"Can only create classes, not {key}")
        if key in self.lookup_stack:
            raise CircularDependency(str(self.lookup_stack))
        info = self._ensure_info(key)
        do_store = kwargs is None and info.store
        combined_kwargs = {**(info.kwargs or {}), **(kwargs or {})}

        inst = self.resolve(info.factory, key=key, repo=repo, kwargs=combined_kwargs)
        if do_store:
            self.add(inst)
        return inst

    def resolve(self, factory, key, repo=None, kwargs=None):
        """Resolve a factory function. This resolves all function parameters as dependencies,
        runs the function and returns the result
        """
        inst = EMPTY
        with self.lookup_stack.push(key):
            for plugin in self.resolvers:
                try:
                    return plugin.create(factory, repo, kwargs)
                except (CannotResolve, PartiallyResolved):
                    continue
            if inst is EMPTY:
                raise CannotResolve(str(self.lookup_stack))

    def _ensure_info(self, cls: t.Type[T]) -> DependencyInfo:
        info = self.types.get(cls)
        if info:
            return info
        self.register(cls)
        return self._ensure_info(cls)

    def add(self, inst, deep=True):
        def append_instance_to(key):
            if key not in self.instances:
                self.instances[key] = []
            self.instances[key].append(inst)

        cls = type(inst)
        self.register(cls)
        append_instance_to(cls)
        if deep:
            for base in cls.__mro__[1:]:
                append_instance_to(base)

    def register(
        self,
        cls: t.Type[T] = None,
        factory: t.Callable = None,
        info: DependencyInfo = None,
        store=True,
        kwargs=None,
    ):
        if not (bool(cls) ^ bool(info)):
            raise ValueError("Supply either cls or info")

        if info is None:
            if not isinstance(cls, type):
                raise TypeError(f"Can only register classes, not {cls}")
            if factory is None:
                factory = cls
            info = DependencyInfo(cls=cls, factory=factory, store=store, kwargs=kwargs)

        for base in info.cls.__mro__:
            if base not in self.types:
                key = base.__name__
                self.types_by_str[key] = base
                self.types[base] = info

    def add_resolver(self, resolver: Resolver):
        self.resolvers.insert(0, resolver)

    def __contains__(self, item: t.Type):
        return item in self.instances


class LayeredRepository(_Stack[SimpleRepository]):
    def __init__(self, first_layer: SimpleRepository):
        super().__init__([first_layer])

    @property
    def current(self) -> SimpleRepository:
        return self[-1]

    def get(self, key: t.Union[t.Type[T], str], many=False, kwargs=None) -> t.Union[t.List[T], T]:
        err = None

        if many:
            return list(chain.from_iterable(repo.get(key, many) for repo in self))

        for repo in reversed(self):
            try:
                return repo.get(key, many, repo=self, kwargs=kwargs)
            except CannotResolve as e:
                err = e
        if err:
            raise err

    def create(self, key: t.Type[T]) -> T:
        return self.current.create(key, repo=self)

    def pop(self):
        if len(self) <= 1:
            raise IndexError("Cannot pop the base repository layer")
        return super().pop()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.pop()
        except IndexError:
            pass

    def __getattr__(self, item):
        return getattr(self.current, item)

    def __contains__(self, item: t.Type):
        return any(item in repo for repo in self)



from __future__ import annotations

import inspect
import typing as t

from gimme.attribute import Attribute
from gimme.exceptions import CannotResolve, PartiallyResolved
from gimme.helpers import parse_type_hint
from gimme.types import T

if t.TYPE_CHECKING:
    from gimme.repository import LayeredRepository


class Resolver:
    """Base class for creating extensions to the dependency resolution system. Subclass this class
    and add instances of your class to the repository using :func:`gimme.add_resolver` or
    :func:`gimme.setup`
    """

    def create(
        self,
        factory: t.Callable,
        repository: LayeredRepository,
        kwargs: dict = None,
    ) -> T:
        kwargs = kwargs or {}
        deps = self.get_dependencies(factory, repository, kwargs)
        deps.update(kwargs)
        return factory(**deps)

    def get_dependencies(
        self,
        factory: t.Callable,
        repository: LayeredRepository,
        kwargs: dict = None,
    ) -> t.Dict[str, t.Any]:
        """Override this to customize how to resolve the dependencies of a specific class /
        factory function. This method will be called with the following arguments:

        :param factory: The class / factory function to determine the dependencies for
        :param repository: The current ``Repository`` can be used for requesting dependencies
        :param kwargs: Any user specified keyword arguments (see :func:`gimme.register`). These
            will have preference over any keyword arguments this function supplies

        Your ``Resolver`` may do one of three things:

        * Return a dictionary of keyword arguments to supply to the factory upon instantiation
        * Raise :exc:`~gimme.exceptions.CannotResolve` if your ``Resolver`` cannot return an
          instantiated object and the next ``Resolver`` should be tried
        * Raise :exc:`~gimme.exceptions.PartiallyResolved` if your ``Resolver`` did resolve some,
          but not all of the dependencies
        """
        raise CannotResolve()


class TypeHintingResolver(Resolver):
    def get_dependencies(
        self,
        factory: t.Callable,
        repository: LayeredRepository,
        kwargs: dict = None,
    ) -> t.Dict[str, t.Any]:
        kwargs = kwargs or {}
        try:
            signature = inspect.signature(factory)
        except ValueError:  # some builtin types
            raise CannotResolve()
        try:
            type_hints = self.get_type_hints(factory, repository)
        except NameError as e:
            raise CannotResolve() from e

        dependencies = {}
        for key, param in signature.parameters.items():
            if key in kwargs or param.default is not inspect.Parameter.empty:
                continue

            try:
                annotation = type_hints[key]
            except KeyError as e:
                raise CannotResolve(key) from e

            if hasattr(annotation, "__origin__"):
                info = parse_type_hint(annotation)
                if info:
                    dependencies[key] = info.collection(repository.get(info.inner_type, many=True))
                    continue
                else:
                    raise CannotResolve(key, annotation)
            dependencies[key] = repository.get(annotation)
        return dependencies

    @staticmethod
    def get_type_hints(obj, repository: LayeredRepository):
        if isinstance(obj, type):
            obj = obj.__init__

        return t.get_type_hints(obj, localns=repository.types_by_str)


class AttributeResolver(Resolver):
    def get_dependencies(
        self,
        factory: t.Callable,
        repository: LayeredRepository,
        kwargs: dict = None,
    ) -> t.Dict[str, t.Any]:
        for attr in vars(factory).values():
            if isinstance(attr, Attribute) and not attr.lazy:
                repository.get(attr.dependency)
        raise PartiallyResolved()
